<?php declare(strict_types=1);

namespace App\Repository;

use App\Controller\DatabaseConnection;
use App\Entity\Nameserver;
use PDO;
use PDOException;

/**
 *
 */
class NameserverRepository
{
	public function __construct(private DatabaseConnection $databaseConnection)
	{
	}
	
	
	/**
	 * @return array
	 */
	public function findAll(): array
	{
		$nameservers = [];
		$sql = "
            SELECT id, name, a, aaaa, apikey
            FROM " . DatabaseConnection::TABLE_NAMESERVERS . "
            ORDER BY name";
		
		try {
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->execute();
			while ($result = $statement->fetch(mode: PDO::FETCH_ASSOC)) {
				$nameserver = new Nameserver(name: $result['name'], id: $result['id'], a: $result['a'], aaaa: $result['aaaa'], apikey: $result['apikey']);
				$nameservers[] = $nameserver;
			}
			return $nameservers;
		} catch (PDOException $e) {
			exit($e->getMessage());
		}
	}
	
	
	/**
	 * @param int $id
	 *
	 * @return \App\Entity\Nameserver
	 */
	public function findByID(int $id): Nameserver
	{
		$sql = "
			SELECT id, name, a, aaaa, apikey
			FROM . " . DatabaseConnection::TABLE_NAMESERVERS . "
            WHERE id = :id";
		
		try {
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->bindParam(param: ':id', var: $id);
			$statement->execute();
			$result = $statement->fetch(mode: PDO::FETCH_ASSOC);
			return new Nameserver(name: $result['name'], a: $result['a'], aaaa: $result['aaaa'], apikey: $result['apikey']);
		} catch (PDOException $e) {
			exit($e->getMessage());
		}
	}
	
	
	/**
	 * @param String $name
	 *
	 * @return \App\Entity\Nameserver|bool
	 */
	public function findByName(string $name): Nameserver|bool
	{
		$sql = "
			SELECT id, name, a, aaaa, apikey
			FROM " . DatabaseConnection::TABLE_NAMESERVERS . "
            WHERE name = :name";
		
		try {
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->bindParam(param: ':name', var: $name);
			$statement->execute();
			if ($result = $statement->fetch(mode: PDO::FETCH_ASSOC)) {
				return new Nameserver(name: $result['name'], a: $result['a'], aaaa: $result['aaaa']);
			} else {
				return false;
			}
		} catch (PDOException $e) {
			exit($e->getMessage());
		}
	}
	
	
	/**
	 * @param \App\Entity\Nameserver $nameserver
	 *
	 * @return string|false
	 */
	public function insert(Nameserver $nameserver): bool|string
	{
		$sql = "
            INSERT INTO " . DatabaseConnection::TABLE_NAMESERVERS . " (name, a, aaaa, apikey)
            VALUES (:name, :a, :aaaa, :apikey)";
		
		try {
			$name = $nameserver->getName();
			$a = $nameserver->getA();
			$aaaa = $nameserver->getAaaa();
			$apikey = $nameserver->getApikey();
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->bindParam(param: ':name', var: $name);
			$statement->bindParam(param: ':a', var: $a);
			$statement->bindParam(param: ':aaaa', var: $aaaa);
			$statement->bindParam(param: ':apikey', var: $apikey);
			$statement->execute();
			
			return $this->databaseConnection->getConnection()->lastInsertId();
		} catch (PDOException $e) {
			exit($e->getMessage());
		}
	}
	
	
	/**
	 * @param Int    $id
	 * @param String $name
	 * @param String $a
	 * @param String $aaaa
	 * @param String $apikey
	 *
	 * @return false|int
	 */
	public function update(int $id, string $name, string $a, string $aaaa, string $apikey): bool|int
	{
		$current = $this->findByID(id: $id);
		
		/* doesn't work
		$statement = "
            INSERT INTO domains(id, name, a, aaaa)
            VALUES(:id, :name, :a, :aaaa)
            ON DUPLICATE KEY UPDATE
            	name=COALESCE(VALUES(name), :name),
				a=COALESCE(:a, a),
            	aaaa=COALESCE(:aaaa, aaaa)";
		*/
		
		if (empty($name)) {
			$name = $current->getName();
		}
		if (empty($a)) {
			$a = $current->getA();
		}
		if (empty($aaaa)) {
			$aaaa = $current->getAaaa();
		}
		if (empty($apikey)) {
			$apikey = $current->getApikey();
		}
		
		$sql = "
			UPDATE " . DatabaseConnection::TABLE_NAMESERVERS . " SET
				name = :name,
				a = :a,
               	aaaa = :aaaa,
               	apikey = :apikey
            WHERE id = :id";
		
		try {
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->bindParam(param: 'id', var: $id);
			$statement->bindParam(param: 'name', var: $name);
			$statement->bindParam(param: 'a', var: $a);
			$statement->bindParam(param: 'aaaa', var: $aaaa);
			$statement->bindParam(param: 'apikey', var: $apikey);
			$statement->execute();
			
			return $statement->rowCount();
		} catch (PDOException $e) {
			echo $e->getMessage();
			return false;
		}
	}
	
	
	/**
	 * @param $id
	 *
	 * @return int
	 */
	public function delete($id): int
	{
		$sql = "
            DELETE FROM " . DatabaseConnection::TABLE_NAMESERVERS . "
            WHERE id = :id";
		
		try {
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->bindParam(param: 'id', var: $id);
			$statement->execute();
			
			return $statement->rowCount();
		} catch (PDOException $e) {
			exit($e->getMessage());
		}
	}
	
	
	/**
	 * @param String $field
	 *
	 * @return int
	 */
	public function getLongestEntry(string $field): int
	{
		$sql = "
            SELECT MAX(LENGTH(" . $field . ")) as length FROM " . DatabaseConnection::TABLE_NAMESERVERS;
		
		try {
			$statement = $this->databaseConnection->getConnection()->prepare(query: $sql);
			$statement->execute();
			$result = $statement->fetch();
			return $result['length'];
		} catch (PDOException $e) {
			exit($e->getMessage());
		}
	}
}